#!/bin/sh
# Time-stamp: <2005.09.27., 10:05:39 (salvi)> \
exec wish "$0" "$@"

# A Tcl/Tk implementation of the board game Pylos.

package require Tk

proc Init {} {
    global images sizes canvasBoard objects gameState
    wm title . Pylos
    wm resizable . 0 0
    
    # loading the images

    set dir [file dirname [info script]]
    foreach i {board red green selected red-small green-small} {
	set images($i) [image create photo -file [file join $dir $i.gif]]
	set sizes($i) [expr [image width $images($i)] / 2.0]
    }
    set boardsize [image width $images(board)]

    # creating the menu and binding the accelerators

    frame .menuBar -borderwidth 0 -takefocus 0
    pack .menuBar -fill x

    menubutton .menuBar.game -text Game -underline 0 -menu .menuBar.game.m
    menubutton .menuBar.help -text Help -underline 0 -menu .menuBar.help.m

    menu .menuBar.game.m -tearoff 0
    .menuBar.game.m add command -label New -underline 0 -accelerator Ctrl+N \
	-command NewGame
    .menuBar.game.m add separator
    .menuBar.game.m add command -label Quit -underline 0 -accelerator Ctrl+Q \
	-command QuitGame
    menu .menuBar.help.m -tearoff 0
    .menuBar.help.m add command -label Rules -underline 0 -command Rules
    .menuBar.help.m add command -label About -underline 0 -command About

    pack .menuBar.game .menuBar.help -side left

    bind . <Control-KeyPress-n> NewGame
    bind . <Control-KeyPress-q> QuitGame
    
    # creating the board with all objects hidden

    set canvasBoard [canvas .canvasBoard -width $boardsize -height $boardsize]
    $canvasBoard create image $sizes(board) $sizes(board) -image $images(board)

    for {set i 0} {$i < 15} {incr i} {
	set objects(redCounter$i) [$canvasBoard create image \
			      [expr $sizes(red-small) + 5] \
			      [expr 15*$i + $sizes(red-small) + 12] \
			      -image $images(red-small) -state hidden]
	set objects(greenCounter$i) [$canvasBoard create image \
				[expr $boardsize - $sizes(green-small) - 6] \
				[expr 15*$i + $sizes(green-small) + 12] \
				-image $images(green-small) -state hidden]
    }

    for {set i 4} {$i > 0} {incr i -1} {
	for {set x 0} {$x < $i} {incr x} {
	    for {set y 0} {$y < $i} {incr y} {
		set objects(piece$i$x$y) [$canvasBoard create image \
				     [expr 25 + $x * 50 + (4 - $i) \
					  * 25 + $sizes(selected)] \
				     [expr 25 + $y * 50 + (4 - $i) \
					  * 25 + $sizes(selected)] \
				     -image $images(selected) -state hidden]
	    }
	}
    }

    pack $canvasBoard

    # setting some variables and binds

    set gameState(playing) 0
    bind . <ButtonPress-1> { HumanPlays place %x %y }
    bind . <Shift-ButtonPress-1> { HumanPlays select %x %y }
    bind . <KeyPress-r> { HumanPlays remove %x %y }
    bind . <KeyPress-R> { HumanPlays remove %x %y }
}

proc NewGame {} {
    global gameState canvasBoard objects    

    if {$gameState(playing)} {
	if {[tk_messageBox -icon warning -title "New Game" -type okcancel \
		 -message "A game is in progress.\nAre you sure?"] \
		eq "cancel"} {
	    return
	}
    }

    # resets the game state
    
    set gameState(playing) 1
    set gameState(actPlayer) red
    set gameState(redRemain) 15
    set gameState(greenRemain) 15
    set gameState(removeMode) 0
    set gameState(selected) none
    set gameState(selected2) none

    # clears the board

    for {set i 4} {$i > 0} {incr i -1} {
	for {set x 0} {$x < $i} {incr x} {
	    for {set y 0} {$y < $i} {incr y} {
		$canvasBoard itemconfigure $objects(piece$i$x$y) -state hidden
	    }
	}
    }
    RedrawRemaining
}

proc QuitGame {} {
    global gameState
    if {$gameState(playing)} {
	if {[tk_messageBox -icon warning -title "New Game" -type okcancel \
		 -message "A game is in progress.\nAre you sure?"] \
		eq "cancel"} {
	    return
	}
    }
    exit
}

proc About {} {
    tk_messageBox -icon info -title "About Pylos" -type ok \
	-message "Pylos\nby Peter Salvi, 2005"
}

proc Rules {} {
    tk_messageBox -icon info -title "Pylos Rules" -type ok \
	-message "Two players play alternately, each has 15 balls. \
                  The board has\
                  16 hollows which allow a pyramid to be built. \
                  The goal is to be the one at the top.\n\nIf\
                  you can get a 'square' formed (4 balls of your\
                  color adjacent to each other on the same level),\
                  you can remove at most 2 of your pieces and put them back\
                  in your reserve.\n\nThere is a piece-saving move\
                  even if the square is not of one color. \
                  Anywhere a square is formed, a player can move a\
                  piece from a lower level, onto the square. \
                  This allows the players to make a play without using\
                  a piece from their reserves.\n\nBalls\
                  can be placed with the left mouse button,\
                  and they can be selected by shift-clicking\
                  (e.g. to move a piece, you should select then\
                  place it). Pressing 'r' removes the selected\
                  ball(s).\n\n\
                  (based on a boardgamegeek review)"
}

# Redraws the markers for the remaining pieces.
proc RedrawRemaining {} {
    global gameState canvasBoard objects
    for {set i 0} {$i < 15} {incr i} {
	if {$gameState(redRemain) > $i} {
	    $canvasBoard itemconfigure $objects(redCounter$i) -state normal
	} else {
	    $canvasBoard itemconfigure $objects(redCounter$i) -state hidden
	}
	if {$gameState(greenRemain) > $i} {
	    $canvasBoard itemconfigure $objects(greenCounter$i) -state normal
	} else {
	    $canvasBoard itemconfigure $objects(greenCounter$i) -state hidden
	}
    }
}

# Handles human play. Type can be place, select or remove.
proc HumanPlays {type x y} {
    global gameState canvasBoard objects images

    if {!$gameState(playing)} {
	return
    }

    if {$type eq "place"} {
	if {$gameState(removeMode)} {
	    return
	}
	set coord [GetEmptyCoord $x $y]
	if {$coord eq "none"} {
	    return
	}
	if {$gameState(selected) eq "none"} {
	    $canvasBoard itemconfigure $objects(piece$coord) -state normal \
		-image $images($gameState(actPlayer))
	    incr gameState($gameState(actPlayer)Remain) -1
	} elseif {[ValidMove $gameState(selected) $coord]} {
	    $canvasBoard itemconfigure $objects(piece$gameState(selected)) \
		-state hidden
	    $canvasBoard itemconfigure $objects(piece$coord) -state normal \
		-image $images($gameState(actPlayer))
	    set gameState(selected) none
	} else {
	    return
	}
	if {[IsSquare $coord]} {
	    set gameState(removeMode) 1
	}
    }

    if {$type eq "select"} {
	set coord [GetCoord $x $y]
	if {$coord eq "none"} {
	    return
	}
	if {[GetPiece $coord] eq "selected"} {
	    $canvasBoard itemconfigure $objects(piece$coord) \
		-image $images($gameState(actPlayer))
	    if {$coord eq $gameState(selected)} {
		set gameState(selected) none
	    } elseif {$coord eq $gameState(selected2)} {
		set gameState(selected2) none
	    }
	    return
	}
	if {$gameState(selected) ne "none"} {
	    if {$gameState(removeMode)} {
		if {$gameState(selected2) ne "none"} {
		    return
		}
		$canvasBoard itemconfigure $objects(piece$coord) \
		    -image $images(selected)
		set gameState(selected2) $coord
		return
	    } else {
		$canvasBoard itemconfigure $objects(piece$gameState(selected))\
		    -image $images($gameState(actPlayer))
	    }
	}
	$canvasBoard itemconfigure $objects(piece$coord) \
	    -image $images(selected)
	set gameState(selected) $coord
	return
    }

    if {$type eq "remove"} {
	if {!$gameState(removeMode)} {
	    return
	}
	if {$gameState(selected) ne "none"} {
	    $canvasBoard itemconfigure $objects(piece$gameState(selected)) \
		-state hidden
	    incr gameState($gameState(actPlayer)Remain)	    
	}
	if {$gameState(selected2) ne "none"} {
	$canvasBoard itemconfigure $objects(piece$gameState(selected2)) \
	    -state hidden
	    incr gameState($gameState(actPlayer)Remain)	    
	}
	set gameState(removeMode) 0
	set gameState(selected) none
	set gameState(selected2) none
    }

    RedrawRemaining

    if {$gameState(removeMode)} {
	return
    }

    if {$gameState(actPlayer) eq "red"} {
	set gameState(actPlayer) green
    } else {
	set gameState(actPlayer) red
    }

    if {$gameState($gameState(actPlayer)Remain) == 0} {
	GameOver $gameState(actPlayer)
    }
}

# Returns the empty ball space ID at the given coordinates,
# or "none" if invalid.
# A valid space:
# - is empty
# - has the given coordinates inside its boundaries
# - is on the first level, or has all four balls under it
proc GetEmptyCoord {mx my} {
    global canvasBoard objects sizes gameState
    for {set i 4} {$i > 0} {incr i -1} {
	for {set x 0} {$x < $i} {incr x} {
	    for {set y 0} {$y < $i} {incr y} {
		if {[GetPiece $i$x$y] ne "none"} {
		    continue
		}
		set tx [expr 25 + $x * 50 + (4 - $i) * 25 + $sizes(selected)]
		set ty [expr 25 + $y * 50 + (4 - $i) * 25 + $sizes(selected)]
		if {$mx < [expr $tx - $sizes($gameState(actPlayer))] ||
		    $mx > [expr $tx + $sizes($gameState(actPlayer))] ||
		    $my < [expr $ty - $sizes($gameState(actPlayer))] ||
		    $my > [expr $ty + $sizes($gameState(actPlayer))]} {
		    continue
		}
		if {$i == 4} {
		    return $i$x$y
		}
		if {[GetPiece [expr $i+1]$x$y] eq "none" ||
		    [GetPiece [expr $i+1][expr $x+1]$y] eq "none" ||
		    [GetPiece [expr $i+1]$x[expr $y+1]] eq "none" ||
		    [GetPiece [expr $i+1][expr $x+1][expr $y+1]] eq "none"} {
		    continue
		}
		return $i$x$y
	    }
	}
    }
    return none
}

# Returns the ball ID at the given coordinates, or none if invalid.
# A valid ball:
# - is of the current player's color or selected
# - has the given coordinates inside its boundaries
# - does not have any balls above it
proc GetCoord {mx my} {
    global canvasBoard objects sizes gameState
    for {set i 4} {$i > 0} {incr i -1} {
	for {set x 0} {$x < $i} {incr x} {
	    for {set y 0} {$y < $i} {incr y} {
		if {[GetPiece $i$x$y] ne $gameState(actPlayer) &&
		    [GetPiece $i$x$y] ne "selected"} {
		    continue
		}
		set tx [expr 25 + $x * 50 + (4 - $i) * 25 + $sizes(selected)]
		set ty [expr 25 + $y * 50 + (4 - $i) * 25 + $sizes(selected)]
		if {$mx < [expr $tx - $sizes($gameState(actPlayer))] ||
		    $mx > [expr $tx + $sizes($gameState(actPlayer))] ||
		    $my < [expr $ty - $sizes($gameState(actPlayer))] ||
		    $my > [expr $ty + $sizes($gameState(actPlayer))]} {
		    continue
		}
		if {$i != [expr $x + 1] && $i != [expr $y + 1] &&
		    [GetPiece [expr $i-1]$x$y] ne "none"} {
		    continue
		}
		if {$i != [expr $y + 1] && $x != 0 &&
		    [GetPiece [expr $i-1][expr $x-1]$y] ne "none"} {
		    continue
		}
		if {$i != [expr $x + 1] && $y != 0 &&
		    [GetPiece [expr $i-1]$x[expr $y-1]] ne "none"} {
		    continue
		}
		if {$x != 0 && $y != 0 &&
		    [GetPiece [expr $i-1][expr $x-1][expr $y-1]] ne "none"} {
		    continue
		}
		return $i$x$y
	    }
	}
    }
    return none
}

# Returns the ball type of the given ID.
# Valid types: none, red, green, selected
proc GetPiece {coord} {
    global canvasBoard objects images
    if {[$canvasBoard itemcget $objects(piece$coord) -state] eq "hidden"} {
	return none
    } elseif {[$canvasBoard itemcget $objects(piece$coord) -image] \
		  eq $images(red)} {
	return red
    } elseif {[$canvasBoard itemcget $objects(piece$coord) -image] \
		  eq $images(green)} {
	return green
    } else {
	return selected
    }
}

proc GameOver {loser} {
    global gameState

    tk_messageBox -icon info -title "Game Over" -type ok \
	-message "Player $loser lost!"
    set gameState(playing) 0
}

proc ValidMove {src dst} {
    set srci [string index $src 0]
    set srcx [string index $src 1]
    set srcy [string index $src 2]
    set dsti [string index $dst 0]
    set dstx [string index $dst 1]
    set dsty [string index $dst 2]

    if {$srci <= $dsti} {
	return false
    }

    if {$srcx == $dstx && $srcy == $dsty} {
	return false
    }
    if {$srcx == [expr $dstx + 1] && $srcy == $dsty} {
	return false
    }
    if {$srcx == $dstx && $srcy == [expr $dsty + 1]} {
	return false
    }
    if {$srcx == [expr $dstx + 1] && $srcy == [expr $dsty + 1]} {
	return false
    }

    return true
}

proc IsSquare {coord} {
    set i [string index $coord 0]
    set x [string index $coord 1]
    set y [string index $coord 2]
    set c [GetPiece $coord]

    if {$x != 0 && $y != 0} {
	if {[GetPiece $i[expr $x - 1][expr $y - 1]] eq $c &&
	    [GetPiece $i[expr $x - 1]$y] eq $c &&
	    [GetPiece $i$x[expr $y - 1]] eq $c} {
	    return true
	}
    }

    if {$x != [expr $i - 1] && $y != 0} {
	if {[GetPiece $i[expr $x + 1][expr $y - 1]] eq $c &&
	    [GetPiece $i[expr $x + 1]$y] eq $c &&
	    [GetPiece $i$x[expr $y - 1]] eq $c} {
	    return true
	}
    }

    if {$x != 0 && $y != [expr $i - 1]} {
	if {[GetPiece $i[expr $x - 1][expr $y + 1]] eq $c &&
	    [GetPiece $i[expr $x - 1]$y] eq $c &&
	    [GetPiece $i$x[expr $y + 1]] eq $c} {
	    return true
	}
    }

    if {$x != [expr $i - 1] && $y != [expr $i - 1]} {
	if {[GetPiece $i[expr $x + 1][expr $y + 1]] eq $c &&
	    [GetPiece $i[expr $x + 1]$y] eq $c &&
	    [GetPiece $i$x[expr $y + 1]] eq $c} {
	    return true
	}
    }

    return false
}

Init
